/*
 * Code to create an account with administrative privileges
 * January 2003
 *
 * Matt Conover (shok@dataforce.net)
 * http://www.w00w00.org
 */

#include "account.h"

extern FILE *debug_file;

// Add an account to a privilege specified by a SID
BOOL AddAccountPrivilege(SID *AccountSID, LPWSTR PrivilegeName)
{
	LSA_HANDLE LSAPolicy;
	LSA_OBJECT_ATTRIBUTES ObjectAttributes;
	LSA_UNICODE_STRING Privilege;
	DWORD ReturnValue, StringLength = wcslen(PrivilegeName);

	memset(&ObjectAttributes, 0, sizeof(ObjectAttributes));
	ReturnValue = LsaNtStatusToWinError(LsaOpenPolicy(NULL, &ObjectAttributes, POLICY_ALL_ACCESS, &LSAPolicy));
	if (ReturnValue != ERROR_SUCCESS)
	{
		if (debug_file) fprintf(debug_file, "LsaOpenPolicy failed: error code 0x%lx\n", ReturnValue);
		return FALSE;
	}

	Privilege.Buffer = PrivilegeName;
	Privilege.Length = (USHORT)(StringLength * sizeof(WCHAR));
	Privilege.MaximumLength = (USHORT)((StringLength+1) * sizeof(WCHAR));
	ReturnValue = LsaNtStatusToWinError(LsaAddAccountRights(LSAPolicy, AccountSID, &Privilege, 1));
	if (ReturnValue != ERROR_SUCCESS)
	{
		if (debug_file) fprintf(debug_file, "LsaAddAccountRights failed: error code 0x%lx\n", ReturnValue);
		return FALSE;
	}

	LsaClose(LSAPolicy);
	return TRUE;
}

// Returns the password to the new account (which the caller need to free) or NULL if it failed
char *AddUserAccount(char *AccountNameA)
{
	DWORD ReturnValue;
	SID *sid;

	/////////////////////////////////////////////////
	// Create a user account
	/////////////////////////////////////////////////

	USER_INFO_1 UserInfo;
	DWORD ErrorIndex = 0;
	WCHAR *AccountNameW;
	DWORD AccountNameLength;
	WCHAR PasswordW[PASSWORD_LEN];
	USER_INFO_1003 UserPassword;
	char *PasswordA;

	AccountNameLength = strlen(AccountNameA) + 1;
	if (!(AccountNameW = (WCHAR *)malloc(AccountNameLength * sizeof(WCHAR)))) return NULL;
	memset(AccountNameW, 0, AccountNameLength);
	MultiByteToWideChar(CP_ACP, 0, AccountNameA, AccountNameLength, AccountNameW, AccountNameLength);

	// Generate a random password and convert it to Unicode
	if (!(PasswordA = (char *)malloc(PASSWORD_LEN))) return NULL;
	memset(PasswordA, 0, PASSWORD_LEN);
	srand(GetTickCount());
	for (register int i = 0; i < PASSWORD_LEN - 1; i++) PasswordA[i] = (rand() % 93) + 0x21;
	MultiByteToWideChar(CP_ACP, 0, PasswordA, PASSWORD_LEN, PasswordW, PASSWORD_LEN);

	UserInfo.usri1_name = AccountNameW;
	UserInfo.usri1_password = PasswordW;
	UserInfo.usri1_priv = USER_PRIV_USER;
	UserInfo.usri1_flags = UF_DONT_EXPIRE_PASSWD | UF_SCRIPT;
	UserInfo.usri1_home_dir = NULL;
	UserInfo.usri1_comment = NULL;
	UserInfo.usri1_script_path = NULL;

	// Create the account
	ReturnValue = NetUserAdd(NULL, 1, (BYTE *)&UserInfo, &ErrorIndex);
	if (ReturnValue != NERR_Success && ReturnValue != NERR_UserExists)
	{
		if (debug_file) fprintf(debug_file, "Error with NetUserAdd: element %d caused error code 0x%lx\n", ErrorIndex, ReturnValue);
		return NULL;
	}

	// Reset the password in case the account already existed
	UserPassword.usri1003_password = PasswordW;
	ReturnValue = NetUserSetInfo(NULL, AccountNameW, 1003, (BYTE *)&UserPassword, &ErrorIndex);
	if (ReturnValue != NERR_Success)
	{
		if (debug_file) fprintf(debug_file, "Error setting password: error code 0x%lx\n", ReturnValue);
		return NULL;
	}

	/////////////////////////////////////////////////
	// Add account to the Administrators group
	/////////////////////////////////////////////////

	LOCALGROUP_MEMBERS_INFO_0 MemberInfo;
	SID_NAME_USE SidType;
	char DomainName[64];
	DWORD SidSize = 0, DomainSize = 0;
	
	// Get SID of the new user
	LookupAccountName(NULL, AccountNameA, NULL, &SidSize, NULL, &DomainSize, &SidType);
	if (DomainSize > sizeof(DomainName)) return NULL;
	if (!(sid = (SID *)malloc(SidSize))) return NULL;
	
	if (!LookupAccountName(NULL, AccountNameA, sid, &SidSize, DomainName, &DomainSize, &SidType))
	{
		if (debug_file) fprintf(debug_file, "LookupAccountName failed: error code 0x%lx\n", GetLastError());
		return NULL;
	}

	MemberInfo.lgrmi0_sid = sid;
	ReturnValue = NetLocalGroupAddMembers(NULL, L"Administrators", 0, (BYTE *)&MemberInfo, 1);
	if (ReturnValue != NERR_Success && ReturnValue != ERROR_MEMBER_IN_ALIAS)
	{
		if (debug_file) fprintf(debug_file, "NetLocalGroupAddMembers failed: error code 0x%lx\n", ReturnValue);
		return NULL;
	}

#ifdef ADD_ACCOUNT_PRIVILEGES // add additional privileges
	if (!AddAccountPrivilege(sid, L"SeBatchLogonRight")) return NULL;
	if (!AddAccountPrivilege(sid, L"SeServiceLogonRight")) return NULL;
	if (!AddAccountPrivilege(sid, L"SeNetworkLogonRight")) return NULL;
	if (!AddAccountPrivilege(sid, L"SeAssignPrimaryTokenPrivilege")) return NULL;
	if (!AddAccountPrivilege(sid, L"SeCreateTokenPrivilege")) return NULL;
	if (!AddAccountPrivilege(sid, L"SeTcbPrivilege")) return NULL;
#endif

	free(sid);
	free(AccountNameW);
	return PasswordA;
}

HANDLE GetTokenThroughAccount(char *AccountNameA, char *PasswordA)
{
	SID *sid;
	SID_NAME_USE SidType;
	char DomainName[64];
	DWORD SidSize = 0, DomainSize = 0;
	HANDLE UserToken;
	
	// Get SID of the new user
	LookupAccountName(NULL, AccountNameA, NULL, &SidSize, NULL, &DomainSize, &SidType);
	if (DomainSize > sizeof(DomainName)) return FALSE;
	if (!(sid = (SID *)malloc(SidSize))) return FALSE;
	
	if (!LookupAccountName(NULL, AccountNameA, sid, &SidSize, DomainName, &DomainSize, &SidType))
	{
		if (debug_file) fprintf(debug_file, "LookupAccountName failed: error code 0x%lx\n", GetLastError());
		return NULL;
	}

	fprintf(debug_file, "Logging in as %s\\%s with the following password: %s\n", DomainName, AccountNameA, PasswordA);
	if (!LogonUser(AccountNameA, DomainName, PasswordA, LOGON32_LOGON_INTERACTIVE, LOGON32_PROVIDER_DEFAULT, &UserToken))
	{
		if (debug_file) fprintf(debug_file, "LogonUser failed: error code 0x%lx\n", GetLastError());
		return NULL;
	}

	return UserToken;
}