/*
 * addr_loop.c
 *
 * Iterate across addresses in a block in a struct addr
 *
 * Copyright (c) 2002 Todd MacDermid <tmacd@synacklabs.net>
 *
 */

#include "addr_loop.h"
#include <sys/types.h>

/* addr_loop will iterate across each address within the netmask provided
 * in the struct addr, in order, calling the callback handler function
 * for each address. The callback handler will receive as arguments
 * a struct addr containing the current address in the iteration, and
 * a void pointer containing whatever got passed to the addr_loop
 * function. If the callback handler ever returns a non-zero value,
 * the loop is terminated. Returns zero upon successful iteration
 * across all addresses.
 */

int addr_loop(struct addr *addr, addr_handler callback, void *arg) {
  struct addr working_addr;
  int ret, addr_size, current_block;
  int overflow = 0;
  uint16_t unmasked_bits;
  uint8_t *mask;
  int i;
  
  if (addr->addr_type == ADDR_TYPE_ETH)
    addr_size = ETH_ADDR_LEN;
  else if (addr->addr_type == ADDR_TYPE_IP)
    addr_size = IP_ADDR_LEN;
  else
    return(-1);
  
  working_addr.addr_type = addr->addr_type;
  working_addr.addr_bits = addr_size * 8;
  
  unmasked_bits = 8*addr_size - addr->addr_bits;
  
  if((mask = (uint8_t *)calloc(1, addr_size)) == NULL) 
    return(-1);
  
  for(i = 0; ((i+1)*8) <= addr->addr_bits; i ++) 
    mask[i] = 0xFF;
  
  
  if((addr->addr_bits % 8) != 0) 
    *(mask + (addr->addr_bits / 8)) = (~0 << (8 - (addr->addr_bits % 8)));
  
  for(i = 0; i < addr_size; i++ ) 
    working_addr.addr_data8[i] = (addr->addr_data8[i] & mask[i]);
  
  while(overflow == 0) {
    if ((ret = (*callback)(&working_addr, arg)) != 0) {
      free(mask);
      return (ret);
    }

    current_block = addr_size - 1;
    while((working_addr.addr_data8[current_block] == 0xFF) && (current_block >= 0)) {
      working_addr.addr_data8[current_block] = 0x00;
      current_block--;
      if(current_block < 0) overflow = 1; /* We're done. */
    }
    working_addr.addr_data8[current_block]++;
    
    for(i = 0; i < addr_size; i++ ) 
      if((working_addr.addr_data8[i] ^ addr->addr_data8[i]) & mask[i]) 
	overflow = 1;
    
  }
  free(mask);
  return(0);
}
  
