/* 
 * lsrscan.c
 *
 * Main code loop
 *
 * Copyright (c) 2002 Todd MacDermid <tmacd@synacklabs.net>
 *
 */

#define FQDNMAX 8192 /* Max length of a domain name */
#define DEFAULTTCPPORT 80

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>

#include "lsrscan.h"
#include "addr_loop.h"
#include "rand.h"
#include "config.h"

extern int h_errno;
extern char *optarg;
extern int optind, opterr, optopt;
const char *verstr= "%s: v0.5 [(c)2002, Todd MacDermid]\n";

void usage(char *programName) {
  printf("usage: %s [-p dstport] [-s srcport] [-S ip] \n\t[-t (to|through|both)] [-b host<:host ...>] [-a host<:host ...>] <hosts ...>\n", programName);
  exit(0);
}

void version(char *programName) {
  printf(verstr, programName);
}

int scan(const struct addr *addr, void *arg) {
  struct settings *settings;
  struct packetInfo packet;
  int extraHops, i;
  int acceptTo = 0;
  int acceptThrough = 0;

  settings = (struct settings *)arg;
  extraHops = settings->numBefore + settings->numAfter;
  packet.hopCount = 3 + extraHops;
  packet.destPort = settings->destPort;
  packet.ipHops[0] = settings->sourceIP;
  packet.sourcePort = settings->sourcePort;

  if((settings->scanType == TYPE_TO) || (settings->scanType == TYPE_BOTH)) {
    if (settings->sourcePort == 0)
      fillBufRandomly(&(packet.sourcePort), 2);
    if (settings->sourceIP == 0) 
      fillBufRandomly(&(packet.ipHops[0]), IP_ADDR_LEN);
    packet.type = TYPE_TO;
    packet.ipHops[2+extraHops] = addr->addr_ip;
    packet.myHop = 1 + settings->numBefore;
    if(settings->silentIP == 0)
      packet.ipHops[packet.myHop] = settings->myIP;
    else
      packet.ipHops[packet.myHop] = settings->silentIP;
    packet.currentHop = 2 + settings->numBefore;
    for(i = 0; i < settings->numBefore; i++) 
      packet.ipHops[1+i] = settings->beforeHops[i];
    for(i = 0; i < settings->numAfter; i++) 
      packet.ipHops[2 + settings->numBefore + i] = settings->afterHops[i];
    if(sendLSR(&packet) < 0) {
      fprintf(stderr,"Error in sendLSR\n");
    } else {
      if(getResponse(settings->pcapPtr, &packet) == 1) 
	printf("%s reverses LSR traffic to it\n", addr_ntoa(addr));
      else
	printf("%s does not reverse LSR traffic to it\n", addr_ntoa(addr));
    }
  }
  if((settings->scanType == TYPE_THROUGH) || 
     (settings->scanType == TYPE_BOTH)) {
    if (settings->sourcePort == 0)
      fillBufRandomly(&(packet.sourcePort), 2);
    packet.type = TYPE_THROUGH;
    packet.myHop = 2+settings->numAfter; /*Gets overwritten*/
    if(settings->silentIP == 0)
      packet.ipHops[2+settings->numAfter] = settings->myIP;
    else
      packet.ipHops[2+settings->numAfter] = settings->silentIP;
    packet.ipHops[1] = addr->addr_ip;
    packet.currentHop = 1;
    for(i = 0; i < settings->numAfter; i++) 
      packet.ipHops[1+i] = 
	settings->afterHops[settings->numAfter-i-1];
    for(i = 0; i < settings->numBefore; i++) 
      packet.ipHops[3 + settings->numAfter + i] =  /*BZZT!*/
	settings->beforeHops[settings->numBefore-i-1];
    if(sendLSR(&packet) < 0) {
      fprintf(stderr,"Error in sendLSR\n");
    } else {
      if(getResponse(settings->pcapPtr, &packet) == 1) 
	printf("%s forwards LSR traffic through it\n", addr_ntoa(addr));
      else
	printf("%s does not forward LSR traffic through it\n", addr_ntoa(addr));
    }
  }
  return(0);
}

char *splitList(char *stringPtr, char *buffer) {
  char *returnPtr;

  memset(buffer, 0, FQDNMAX);
  returnPtr = strchr(stringPtr, 58);
  if (returnPtr == NULL) {
    strncpy(buffer, stringPtr, FQDNMAX-1);
  } else {
    memcpy(buffer, stringPtr, (returnPtr - stringPtr));
    returnPtr++;
  }
  return(returnPtr);
}

int main(int argc, char **argv) {
  struct settings settings;
  struct addr tmpAddr;
  int c, i, tmpPort;
  char tempName[FQDNMAX];
  char *tempNamePtr;

  if (argc < 2) usage(argv[0]);

  if(!(getuid() == 0)) {
    printf("Gotta be root to run %s\n", argv[0]);
    return(-1);
  }

  /*
   * s: source port (default random)
   * S: source IP address (default random)
   * p: destination port (default 80)
   * i: "silent" IP. i.e. you're not using your real IP, but one you
   *    can sniff anyways.
   * t: Type of scan, to or through or both. Default both
   * a: Hosts "after" your current host on the source routed hops. These
   *    hosts must forward source routed packets for lsrscan to work
   * b: Hosts "before" your current host on the source routed hops. These
   *    hosts do not need to support source routing, as they will never 
   *    see packets.
   * v: verbose mode
   * V: Display version number.
   */

  settings.silentIP = 0;
  settings.destPort = DEFAULTTCPPORT;
  settings.scanType = TYPE_BOTH;
  settings.sourceIP = 0;
  settings.sourcePort = 0;
  settings.verbose = 0;
  settings.numBefore = 0;
  settings.numAfter = 0;

  while((c=getopt(argc, argv, "S:s:p:i:b:a:t:hvV")) != -1) {
    switch (c) {
    case 'S':
      if ((addr_pton(optarg, &tmpAddr)) < 0) {
	fprintf(stderr, "Cannot resolve source IP: %s\n", optarg);
	return(-1);
      }
      settings.sourceIP = tmpAddr.addr_ip;
      break;

    case 'h':
      usage(argv[0]);
      break;      
      
    case 'v':
      settings.verbose = 1;
      break;
      
    case 'V':
      version(argv[0]);
      break;
      
    case 's':
      tmpPort = atoi(optarg);
      if((tmpPort < 0) || (tmpPort > TCP_PORT_MAX)) {
	fprintf(stderr, "Invalid source port: %s\n", optarg);
	return(-1);
      }
      settings.sourcePort = tmpPort;
      break;

    case 'p':
      tmpPort = atoi(optarg);
      if((tmpPort < 0) || (tmpPort > TCP_PORT_MAX)) {
	fprintf(stderr, "Invalid dest port: %s\n", optarg);
	return(-1);
      }
      settings.destPort = tmpPort;
      break;
      
    case 'i':
      if ((addr_pton(optarg, &tmpAddr)) < 0) {
	fprintf(stderr, "Cannot resolve silent IP: %s\n", optarg);
	return(-1);
      } 
      settings.silentIP = tmpAddr.addr_ip;
      break;

    case 't':
      if(strstr(optarg, "to")) settings.scanType = TYPE_TO; 
      else if (strstr(optarg, "b")) settings.scanType = TYPE_BOTH;
      else if (strstr(optarg, "th")) settings.scanType = TYPE_THROUGH;
      else {
	fprintf(stderr, "Invalid scan type. valid:(to|through|both)\n");
	return(-1);
      }
      break;

    case 'b':
      tempNamePtr = optarg;
      do {
	(tempNamePtr = splitList(tempNamePtr, tempName));
	if ((addr_pton(tempName, &tmpAddr)) < 0) {
	  fprintf(stderr, "Cannot resolve host: %s\n", tempName);
	  return(-1);
	}
	settings.beforeHops[settings.numBefore] = tmpAddr.addr_ip;
	settings.numBefore++;
      } while((tempNamePtr != NULL) && (settings.numBefore <= 8));
      if (settings.numBefore >= 8) {
	fprintf(stderr, "Too many hosts listed in before group\n");
	return(-1);
      }
      break;
      
    case 'a':
      tempNamePtr = optarg;
      do {
	(tempNamePtr = splitList(tempNamePtr, tempName));
	if ((addr_pton(tempName, &tmpAddr)) < 0) {
	  fprintf(stderr, "Cannot resolve host: %s\n", tempName);
	  return(-1);
	}
	settings.afterHops[settings.numAfter] = tmpAddr.addr_ip;
	settings.numAfter++;
      } while((tempNamePtr != NULL) && (settings.numAfter <= 8));
      if (settings.numAfter >= 8) {
	fprintf(stderr, "Too many hosts listed in after group\n");
	return(-1);
      }
      break;
      
    }
  }
  
  if(settings.numBefore + settings.numAfter > 8) {
    fprintf(stderr, "Maximum of 8 source routed hosts\n");
    return(-1);
  }

  if((settings.pcapPtr = initPcap(&settings)) == NULL) 
    return(-1); /* I think I've already got error messages on all the ways 
		 * you can bomb out */
  
  for(i = optind; i < argc; i++) {
    if ((addr_pton(argv[i], &tmpAddr)) < 0) {
      fprintf(stderr, "Cannot resolve dest IP: %s\n", optarg);
    } else {
      addr_loop(&tmpAddr, scan, &settings);
    }
  }
  pcap_close(settings.pcapPtr);
}
