/*
 * packetsend.c
 *
 * Function to build and send probe packets.
 *
 * Copyright (c) 2002 Todd MacDermid <tmacd@synacklabs.net>
 *
 */

#define LSROVERHEAD 3 /* Number of bytes used by an empty LSR option */

#include "lsrscan.h"
#include "rand.h"
#include "config.h"

/* fillIPHeader - fill in all IP header fields, except for options */

int fillIPHeader(uint8_t *headerBuf, struct packetInfo *packet) {
  struct ip_hdr *ipHeader;
  int optionSize;
  int optionWords;
  uint16_t length;

  optionSize = LSROVERHEAD+(packet->hopCount-2)*IP_ADDR_LEN;
  if (optionSize%4)
    optionSize += (4-(optionSize%4)); /* Pad out to word boundary */
  optionWords = optionSize / 4;

  length = IP_HDR_LEN + optionWords*4 + TCP_HDR_LEN;
  
  ipHeader = (struct ip_hdr *)headerBuf;
  ipHeader->ip_v = 4;
  ipHeader->ip_hl = 5 + optionWords;
  ipHeader->ip_len = htons(length);
  fillBufRandomly(&(ipHeader->ip_id), 2);
  ipHeader->ip_ttl = 255;
  ipHeader->ip_p = IPPROTO_TCP;
  ipHeader->ip_src = packet->ipHops[0];
  ipHeader->ip_dst = packet->ipHops[packet->currentHop];
  
  return(0);
}

/* fillIPOptLSR - Fill in the IP options field with a loose source route
 * option containing all requested hops
 */

/* remember that the current destination IP address and the
 * source address are not in the LSR field. See TCP/IP vol2
 * page 255 
 */

int fillIPOptLSR(uint8_t *buffer, struct packetInfo *packet){
  uint8_t i;
  
  *buffer = IP_OPT_LSRR;
  buffer++;
  *buffer = LSROVERHEAD+(IP_ADDR_LEN*(packet->hopCount - 2));
  buffer++;
  *buffer = (packet->currentHop*IP_ADDR_LEN);
  buffer++;
  
  for(i = 1; i < (packet->hopCount); i++) {
    if (!(i == packet->currentHop)) {
      memcpy (buffer, &(packet->ipHops[i]), IP_ADDR_LEN);
      buffer+=IP_ADDR_LEN;
    }
  }
  return(0);
}

/* fillTCPHeader - fill in TCP header fields */

int fillTCPHeader(uint8_t *buffer, struct packetInfo *packet) {
   struct tcp_hdr *tcpHeader;
  
  tcpHeader = (struct tcp_hdr *)buffer;
  tcpHeader->th_sport = htons(packet->sourcePort);
  tcpHeader->th_dport = htons(packet->destPort);
  fillBufRandomly(&(tcpHeader->th_seq), 4);
  packet->sequenceNum = tcpHeader->th_seq;
  tcpHeader->th_off = 5;
  tcpHeader->th_flags = TH_SYN;
  tcpHeader->th_win = htons(31856); /* FIX ME!!! Do I care about this? */
  
  return(0);
}

/* SendLSR - Opens an ip handle, constructs the actual packet, and
 * sends it. Returns 0 on success, -1 on failure.
 */

int sendLSR(struct packetInfo *packet) { 
  ip_t *ip;
  uint8_t *packetBuf;
  uint16_t packetSize;
  struct tcp_hdr *tcpHeader;
  int IPoptionSize;
  
  if((ip = ip_open()) == NULL) {
    fprintf(stderr, 
	    "Cannot initialize socket (not root or unsupported interface)\n");
    return(-1);
  }
  
  IPoptionSize = LSROVERHEAD + (packet->hopCount-2)*IP_ADDR_LEN;
  if (IPoptionSize%4)
    IPoptionSize += (4-(IPoptionSize%4)); /* Pad out to word boundary */
  if(IPoptionSize > IP_OPT_LEN_MAX) {
    fprintf(stderr, "sendTCPLSR: Options too large\n");
    return(-1); 
  }
  
  packetSize = IP_HDR_LEN + TCP_HDR_LEN + IPoptionSize;
  if((packetBuf = (uint8_t *)calloc(1,packetSize)) == NULL) {
    fprintf(stderr, "sendLSR: calloc returned NULL\n");
    return(-1);
  }

  fillIPHeader(packetBuf, packet);
  fillIPOptLSR(packetBuf+IP_HDR_LEN, packet);
  fillTCPHeader(packetBuf+IP_HDR_LEN+IPoptionSize, packet);

  ip_checksum(packetBuf, packetSize);
  
  ip_send(ip, packetBuf, packetSize);

  ip_close(ip);
  
  free(packetBuf);
  
  return(0);
}
