/*
 * pcapf.c
 *
 * Contains functions to interface with libpcap
 *
 * Copyright (c) 2002 Todd MacDermid <tmacd@synacklabs.net>
 *
 */

#define MAXPACKET 65535
#define PACKETWAIT 10
#define FILTERSIZE 256
#define IPSTRINGSIZE 20
#define TMPBUFSIZE 1024

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "lsrscan.h"

/* getLinkSize - Return the length of the link layer header on packets */

int getLinkSize(pcap_t *pcapPtr) {
  int linkType;
  
  linkType = pcap_datalink(pcapPtr);
  
  if(linkType == DLT_NULL) return(4);
  else if (linkType == DLT_EN10MB) return(14);
  else if (linkType == DLT_IEEE802) return(22);
  else if (linkType == DLT_RAW) return (0);
  else if (linkType == DLT_FDDI) return (21);
  else return(-1); /* Unsupported as of yet...*/
} 

/* local_ntop - Just a silly wrapper, because libpcap doesn't speak in
 * struct addrs, but I still prefer addr_ntop to inet_ntoa.
 */

int local_ntop(uint32_t addr, char *string, int stringSize) {
  struct addr tmpAddr;

  memcpy(&(tmpAddr.addr_ip), &addr, IP_ADDR_LEN);
  tmpAddr.addr_bits = 32;
  tmpAddr.addr_type = ADDR_TYPE_IP;
  addr_ntop(&tmpAddr, string, stringSize);
  return(0);
}

/* GetIntIP - Returns an IP address for the interface passed in. Uses
 * dnet capabilities to do this, because pcap_lookupnet only gets the
 * network number. Because this function returns an unsigned int, it
 * returns 0 on FAILURE, and an IP address on success.
 */

uint32_t getIntIP (char *interface) {
  uint8_t *buf;
  intf_t *intHandle;
  struct intf_entry *intfEntry;
  uint32_t addr;
  
  if((intHandle=intf_open()) == NULL) {
    fprintf(stderr, "getIntIP: intf_open failed\n");
    return(0);
  }
  if((buf=(uint8_t *)calloc(1, TMPBUFSIZE)) == NULL) {
    fprintf(stderr, "getIntIP: calloc failed\n");
    return(0);
  }
  
  intfEntry = (struct intf_entry *)buf; 
  strncpy(intfEntry->intf_name, interface, 59);
  intfEntry->intf_len = TMPBUFSIZE;
  
  if(intf_get(intHandle, intfEntry) < 0) {
    fprintf(stderr, "getIntIP: intf_get failed\n");
    intf_close(intHandle);
    free(buf);
    return(0);
  }

  addr = intfEntry->intf_addr.addr_ip;
  free(buf);
  return(addr);
}

/* setFilter - Compile and set the filter code to be used on the pcap
 * structure. Return 0 on success, -1 on failure.
 */

int setFilter(pcap_t *pcapPtr, char *filter) {
  char errbuf[PCAP_ERRBUF_SIZE];
  struct bpf_program filterCode;
  bpf_u_int32 myAddr, myNetmask;
  int ret;
 
  if ((pcap_compile(pcapPtr, &filterCode, filter, 1, 0)) < 0) {
    fprintf(stderr, "pcap_compile failed: %s\n", pcap_geterr(pcapPtr));
    return(-1);
  }

  ret = pcap_setfilter(pcapPtr, &filterCode);
  pcap_freecode(&filterCode);

  if(ret < 0)
    fprintf(stderr, "pcap_setfilter failed: %s\n", pcap_geterr(pcapPtr)); 

  return (ret);
}

/* openPcap -  open a pcap structure, and set it into nonblocking mode */

pcap_t * openPcap(char *interface, int promiscuous) {
  char errbuf[PCAP_ERRBUF_SIZE];
  pcap_t *pcapPtr;
  
  if((pcapPtr = pcap_open_live(interface, MAXPACKET, promiscuous, PACKETWAIT,
			       errbuf)) == NULL) {
    fprintf(stderr, "pcap_open_live failed: %s\n",errbuf);
    return(NULL);
  }

  if((pcap_setnonblock(pcapPtr, 1, errbuf)) < 0) {
    fprintf(stderr, "pcap_setnonblock failed: %s\n",errbuf);
    pcap_close(pcapPtr);
    return(NULL);
  } 
  
  return (pcapPtr);
}
  
pcap_t * initPcap(struct settings *settings) {
  pcap_t *pcapPtr;
  char *interface;
  bpf_u_int32 myAddr, myNetmask;
  char filter[FILTERSIZE];
  char myIPString[IPSTRINGSIZE];
  char errbuf[PCAP_ERRBUF_SIZE];

  if((interface = pcap_lookupdev(errbuf)) == NULL) {
    fprintf(stderr, "pcap_lookupdev failed: %s\n", errbuf);
    return(NULL);
  }

  if((pcap_lookupnet(interface, &myAddr, &myNetmask, errbuf)) < 0) {
    fprintf(stderr, "pcap_lookupnet failed: %s\n", errbuf);
    return(NULL);
  }


  settings->myIP = getIntIP(interface);
  if(settings->myIP == 0)
    return(NULL);

  if(settings->silentIP) {
    pcapPtr = openPcap(interface, 1);
    if((local_ntop(settings->silentIP, myIPString, IPSTRINGSIZE)) < 0 ) {
     fprintf(stderr, "addr_ntop: listenAddr conversion failed\n");
      return(NULL);
    }
  } else {
    pcapPtr = openPcap(interface, 0);
    if((local_ntop(settings->myIP, myIPString, IPSTRINGSIZE)) < 0 ) {
      fprintf(stderr, "addr_ntop: listenAddr conversion failed\n");
      return(NULL);
    }
  }

  if(pcapPtr == NULL)
    return(NULL);
  
  snprintf(filter, sizeof(filter)-1, "tcp and port %d and host %s",
	   settings->destPort, myIPString);
  
  if(setFilter(pcapPtr, filter) < 0) {
    pcap_close(pcapPtr);
    return(NULL);
  }
  return(pcapPtr);
}
