/*
 * forkarp.c
 *
 * Copyright (c) 2002 Todd MacDermid <tmacd@synacklabs.net>
 *
 */

#define PACKETSIZE 1518 /* No, I don't expect to get that big */
#define PACKETWAIT 300 /* In milliseconds */
#define FILTERSIZE 512

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <netinet/in.h>
#include <dnet.h>
#include <pcap.h>

arp_t *arpHandle;
struct arp_entry arpEntry;
int unsignaled = 1;

void die (char *message1, char *message2) {
  fprintf(stderr, "%s %s\n", message1, message2);
  exit(-1);
}

void removeArpEntry() {
  arp_delete(arpHandle, &arpEntry);
  arp_close(arpHandle);
}

void endProc(int throwaway) {
  exit(0);
}

int forkArpd(struct addr *silentAddr) {
  pid_t pid;
  uint8_t *inPacket;
  uint8_t outPacket[PACKETSIZE];
  uint8_t ipAddrMask[4];
  struct arp_hdr *inArpHeader, *outArpHeader;
  struct arp_ethip *inArpMessage, *outArpMessage;
  struct eth_hdr *outEthernet;
  char errbuf[PCAP_ERRBUF_SIZE];
  char filter[FILTERSIZE];
  struct bpf_program filterCode;
  struct pcap_pkthdr packetHead;
  pcap_t *pcapPtr;
  bpf_u_int32 myAddr, myNetmask;
  char *interface;
  eth_t *ethernetHandle;
  eth_addr_t myMAC;
  int match, ret, i;
  int linkHeadSize = 14;

  pid = fork();
  if (pid != 0) 
    return(pid);
  
  for(i = 0; ((i+1)*8) <= silentAddr->addr_bits; i ++) 
    ipAddrMask[i] = 0xFF;
  if((silentAddr->addr_bits % 8) != 0) 
    *(ipAddrMask + (silentAddr->addr_bits / 8)) = (~0 << (8 - (silentAddr->addr_bits % 8)));

  if((interface = pcap_lookupdev(errbuf)) == NULL) 
    die("pcap_lookupdev failed:", errbuf);

  if((pcap_lookupnet(interface, &myAddr, &myNetmask, errbuf)) < 0)
    die("pcap_lookupnet failed:", errbuf);

  if((pcapPtr = pcap_open_live(interface, PACKETSIZE, 1, PACKETWAIT,
			       errbuf)) == NULL) 
    die("pcap_open_live failed:", errbuf);
  
  snprintf(filter, sizeof(filter)-1, "arp");
  if ((pcap_compile(pcapPtr, &filterCode, filter, 1, myNetmask)) < 0) 
    die("pcap_compile failed:", errbuf);

  ret = pcap_setfilter(pcapPtr, &filterCode);
  pcap_freecode(&filterCode);
  if(ret < 0) 
    die("pcap_setfilter failed:", pcap_geterr(pcapPtr));
  
  /*  if((linkHeadSize = getLinkSize(pcapPtr)) < 0) 
      die("Unsupported link type", NULL); */

  if ((ethernetHandle=eth_open(interface)) == NULL)
    die("Unable to open ethernet interface", NULL);

  if (eth_get(ethernetHandle, &myMAC) < 0)
    die("Unable to obtain hardware address", NULL);

  if ((arpHandle = arp_open()) == NULL)
    die("Unable to open arp cache", NULL);

  memcpy(&(arpEntry.arp_pa), silentAddr, sizeof(struct addr));
  memcpy(&(arpEntry.arp_ha.addr_eth), &myMAC, ETH_ADDR_LEN);
  arpEntry.arp_ha.addr_type = ADDR_TYPE_ETH;
  arpEntry.arp_ha.addr_bits = ETH_ADDR_LEN * 8;
  if ((arp_add(arpHandle, &arpEntry)) < 0)
    die("Error adding entry to arp table", NULL);

  atexit(removeArpEntry);
  signal(SIGINT, endProc);

  while(1) {
    inPacket = NULL;
    while((inPacket = (uint8_t *)pcap_next(pcapPtr, &packetHead)) != NULL) {

      inPacket += linkHeadSize;
      inArpHeader = (struct arp_hdr *)inPacket;
      if(inArpHeader->ar_op == htons(ARP_OP_REQUEST)) {
	inArpMessage = (struct arp_ethip *)(inPacket + sizeof(struct arp_hdr));
	match = 1;
	for(i = 0; i < IP_ADDR_LEN; i++) {
	  if (!(((inArpMessage->ar_tpa)[i] & ipAddrMask[i]) ==
	      ((silentAddr->addr_data8)[i] & ipAddrMask[i]))) {
	    match = 0;
	  }
	}
	if(match) {
	  memset(outPacket, 0x00, PACKETSIZE);
	  outEthernet = (struct eth_hdr *)outPacket;
	  memcpy(outEthernet->eth_dst.data, inArpMessage->ar_sha, ETH_ADDR_LEN);
	  memcpy(outEthernet->eth_src.data, &myMAC, ETH_ADDR_LEN);
	  
	  outEthernet->eth_type = htons(ETH_TYPE_ARP);
	  outArpHeader = (struct arp_hdr *)(outPacket + sizeof(struct eth_hdr));
	  outArpHeader->ar_hrd = htons(ARP_HRD_ETH);
	  outArpHeader->ar_pro = htons(ARP_PRO_IP);
	  outArpHeader->ar_hln = ETH_ADDR_LEN;
	  outArpHeader->ar_pln = IP_ADDR_LEN;
	  outArpHeader->ar_op = htons(ARP_OP_REPLY);
	  outArpMessage = (struct arp_ethip *)((uint8_t *)outArpHeader + sizeof(struct arp_hdr));

	  memcpy(&(outArpMessage->ar_sha), &myMAC, ETH_ADDR_LEN);
	  memcpy(&(outArpMessage->ar_spa), &(inArpMessage->ar_tpa), IP_ADDR_LEN);
	  memcpy(&(outArpMessage->ar_tha), &(inArpMessage->ar_sha), ETH_ADDR_LEN);
	  memcpy(&(outArpMessage->ar_tpa), &(inArpMessage->ar_spa), IP_ADDR_LEN); 
	  eth_send(ethernetHandle, outPacket, 
		   sizeof(struct eth_hdr) + sizeof(struct arp_hdr) +
		   2*IP_ADDR_LEN + 2* ETH_ADDR_LEN); 
	}
      }
    }
  }
}

