/**
 * $Id: bintree.h,v 1.8 2003/02/25 16:37:28 plasmahh Exp $
 * $Revision: 1.8 $
 * $Author: plasmahh $
 * $Date: 2003/02/25 16:37:28 $
 */

#ifndef __EXCEPTION_H
#include "exception.h"
#endif

/**
 * This is the Base class for a binary tree. It implements search and simple
 * add functions, as well as remove by key.
 */
template <class Key,class Data>
class treenode;
template <class Key,class Data>
class bintreeiterator;
template <class Key,class Data>
class bintree /*{{{*/
{
friend class bintreeiterator<Key,Data>;
public:
    /**
     * Adds a copy of the referenced Key/Data pair into the tree, in usual
     * bintree way (i.e. does not reorder/rebal)
     * @param Key& Reference to the key, that will be used to order the data
     * into the tree
     * @param Data& The Data that will be stored into the tree
     * @return none
     * @note This function is virtual, since a derived class may use another
     * way to order the elements (e.g. an avl-tree).
     * @note If the Key is already in the tree, it will be overwritten (member
     * to member copy)
     */
    virtual void add ( const Key& nk, const Data& nd );
    
    /**
     * Removes the node with the given key. The Data and key will be internally deleted.
     * @param Key& Reference to the Key, that should be removed from the tree.
     * @return bool true, if the key could be found & deleted, false otherwise.
     */
    virtual bool remove( const Key& dk);
    
    /**
     * Gets some Data by Key.
     * @param Key& Reference to the Key, that should be searched for in the tree.
     * @return Data& Reference to the Data that has been found from the key.
     * @note If there is no such Key, the tree will throw an exception pointer,
     * that need to be catched and handled. If speed does not matter, you may
     * even use the has_key function, that will do most likely the same.
     */
    Data& get_key(const Key& sk);
    
    /**
     * Looks, if the Key is within the tree.
     * @param Key& Reference to the key, that we should look for.
     * @return bool if the key is in the tree, false otherwise
     */
    bool  has_key(const Key& sk) const;

    inline uint32 size( void ) const { return items; }
    inline bool empty( void )  const { return (items == 0); }

    void clear( void );
    bintree( void ) : Root(NULL),items(0) { }
    bintree( const bintree& obt);
    virtual ~bintree( void ) { clear(); }
private :
    treenode<Key,Data> * Root;
    uint32 items;
protected:
    inline void printon_sub(treenode<Key,Data>* pn, ostream & o,mstring prefix,const mstring& direction) const;
    
friend ostream& operator<< <Key,Data> ( ostream & o, const bintree<Key,Data> & t);
};/*}}}*/
template <class Key,class Data>
bintree<Key,Data>::bintree( const bintree<Key,Data> & obt) : Root(NULL), items(0)
{  
    bintreeiterator<Key,Data> bit(obt);
    for ( bit.first(); bit.valid(); bit.next())
    {
	add ( bit.key(), bit.data() );
    }
}
template <class Key,class Data>
inline void bintree<Key,Data>::printon_sub(treenode<Key,Data>* pn,ostream & o,mstring prefix,const mstring& direction) const/*{{{*/
{
    mstring nprefix;
    o << direction << *pn << endl;
    
    if ( pn->left && pn->right )
    {
        o << prefix << "|---";
	nprefix = prefix + "|    ";
    }
    else if ( !pn->left && !pn->right )
    {
    }
    else 
    {
	o << prefix << "`---";
	nprefix = prefix + "     ";
    }
    
    if ( pn->left )
    {
	printon_sub(pn->left,o,nprefix," [L]");
	if ( pn->right )
	{
	    o << prefix << "`---";
	}
    }
    if ( pn->right )
    {
	nprefix = prefix + "     ";
	printon_sub(pn->right,o,nprefix," [R]");
    }
}/*}}}*/
template <class Key,class Data>
ostream& operator<< ( ostream & o, const bintree<Key,Data> & t)/*{{{*/
{
    if (!t.Root) 
    {
	o << "(empty)";
    }
    else
    {
	o << "  ";
	t.printon_sub(t.Root,o,"  ","");
	o << "Size : " << t.items << endl;
    }
    return o;
}/*}}}*/
template <class Key,class Data>
void bintree<Key,Data>::add( const Key& nk, const Data& nd)/*{{{*/
{
    treenode<Key,Data> * nn = new treenode<Key,Data>(nk,nd);

    if (!Root)
    {
	Root = nn;
	items = 1;
	return;
    }

    treenode<Key,Data> * ptr = Root;
    while ( true ) //ptr->left && ptr->right )
    {
	if ( nn->_key < ptr->_key )
	{
	    // kleiner als der knoten, also weiter links im baum.  wenns nicht
	    // weiter links geht, dann hier einfgen.  ansonsten ptr weiter
	    // links setzen...
	    if ( ptr->left )
	    {
		ptr = ptr->left;
		continue;
	    } 
	    // ok, links ist kein knoten, also an den linken knoten anfgen...
	    ptr->left = nn;
	    nn->parent = ptr;
	    items++;
	    return;
	}	
	else if ( ptr->_key < nn->_key )
	{
	    // grsser als der knoten, also rechts im baum.
	    if ( ptr->right )
	    {
		ptr = ptr->right;
		continue;
	    }
	    ptr->right = nn;
	    nn->parent = ptr;
	    items++;
	    return;
	}
	else
	{
	    // ok, muss also gleich sein, daher den knoten ersetzen ???
	    ptr->_data = nn->_data;
	    ptr->_key  = nn->_key;
	    delete nn;
	    return;
	}
    }
    // ok, wenn wir hier sind, dann wurde noch nichts aufgrund teilweise besetzter
    // knoten eingefgt denke ich mal...
}/*}}}*/
template <class Key,class Data>
bool bintree<Key,Data>::remove( const Key& rk)/*{{{*/
{
    if (!Root) return false;
    // Do a search, but remember the node the found key is bound to !!
    treenode<Key,Data> * ptr = Root;

    while (ptr)
    {
	if ( rk < ptr->_key ) // key is kleiner als der knoten, also gehe nach links weiter.
	{
	    //cout << "GOING FROM " << *ptr << " TO " << *ptr->left << endl;
	    ptr = ptr->left;
	    continue;
	}
	else if ( ptr->_key < rk)
	{
	    //cout << "GOING FROM " << *ptr << " TO " << *ptr->right << endl;
	    ptr = ptr->right;
	    continue;
	}
	else
	{
	    cout << "FOUND" << endl;
	    // ok, ist also gleich, d.h. removen.
	    // dazu suche ich zuerst im rechten teilbaum den kleinsten
	    // (usserst linken) knoten heraus, und ersetze den zu entfernenden
	    // durch diesen. Normalerweise msste ich den grsseren von beiden
	    // teilbumen nehmen, jedoch verbraucht das wieder ne menge zeit.
	    if ( ! ptr->right )
	    {
		if (ptr->parent)
		{
		    if (ptr == ptr->parent->left)
		    {
			ptr->parent->left = ptr->left;
		    }
		    else
		    {
			ptr->parent->right = ptr->left;
		    }
		}
		else
		{
		    Root = ptr->left;
		}
		if (ptr->left)
		{
		    ptr->left->parent = ptr->parent;
		}
		items--;
		delete ptr;
		return true;
	    }
	    if ( ! ptr->left )
	    {
		if (ptr->parent)
		{
		    if (ptr == ptr->parent->left)
		    {
			ptr->parent->left = ptr->right;
		    }
		    else
		    {
			ptr->parent->right = ptr->right;
		    }
		}
		else
		{
		    Root = ptr->left;
		}
		if (ptr->right)
		{
		    ptr->right->parent = ptr->parent;
		}
		items--;
		delete ptr;
		return true;
	    }

	    treenode<Key,Data> * ssm = ptr->right;
	    while ( ssm->left )
	    {
		ssm = ssm->left;
	    }

	    ptr->left->parent = ssm;
	    if ( ssm == ssm->parent->left )
	    {
		ssm->parent->left = ssm->right;
	    }
	    else
	    {
		ssm->parent->right = ssm->right;
	    }
	    if (ssm->right)
	    {
		ssm->right->parent = ssm->parent;
	    }
	    
	    ssm->left = ptr->left;
	    ssm->right = ptr->right;
	    ssm->parent = ptr->parent;
	    if ( ptr == ptr->parent->left )
	    {
		ptr->parent->left = ssm;
	    }
	    else
	    {
		ptr->parent->right = ssm;
	    }
	    
	    items--;
	    delete ptr;
	    return true;
	}
    }
    return false;

}/*}}}*/
template <class Key,class Data>
Data& bintree<Key,Data>::get_key( const Key& sk)/*{{{*/
{
    treenode<Key,Data> * ptr = Root;

    if (!Root)
    {
	throw x_key("bintree::get_key : tree is empty", __FLF__);
    }

    while ( ptr )
    {
	if ( ptr->_key < sk )
	{
	    ptr = ptr->right;
	    continue;
	}
	if ( sk < ptr->_key )
	{
	    ptr = ptr->left;
	    continue;
	}
	return ptr->_data;
    }
    throw x_key("bintree::get_key : Key Not Found",__FLF__);
}/*}}}*/
template <class Key,class Data>
bool  bintree<Key,Data>::has_key(const Key& sk) const/*{{{*/
{
    treenode<Key,Data> * ptr = Root;

    if (!Root)
    {
	return false;
    }

    while ( ptr )
    {
	if ( ptr->_key < sk )
	{
	    ptr = ptr->right;
	    continue;
	}
	if ( sk < ptr->_key )
	{
	    ptr = ptr->left;
	    continue;
	}
	return true;
    }
    return false;
}/*}}}*/
template <class Key,class Data>
void bintree<Key,Data>::clear( void )/*{{{*/
{
    if (!Root) return;

    treenode<Key,Data> * ptr = Root;
    treenode<Key,Data> * dn  = Root;

    while (ptr)
    {

	if (ptr->left)
	{
	    ptr = ptr->left;
	    if ( ptr->parent )
	    {
		ptr->parent->left = NULL;
	    }
	    continue;
	}

	if (ptr->right)
	{
	    ptr = ptr->right;
	    if ( ptr->parent )
	    {
		ptr->parent->right = NULL;
	    }
	    continue;
	}
	dn = ptr;
	ptr = ptr->parent;
	delete dn;
    }
    Root = NULL;
}/*}}}*/
enum avl_tree_skew { left, right, noone };
template <class Key,class Data>
class treenode/*{{{*/
{
friend class bintree<Key,Data>;
friend class bintreeiterator<Key,Data>;
public:
    /// accesses the current item
    inline operator       Data&( )       { return _data; };
    /// accesses the current item via const reference
    inline operator const Data&( ) const { return _data; };
    /// accesses the current item via pointer reference
    inline operator const Data*( ) const { return & _data; };
    /// Returns the Next Element of this Node
    inline treenode<Key,Data> *Left( void ) const { return left; }
    /// Returns Previous Element to this Node
    inline treenode<Key,Data> *Right( void ) const { return right; }
protected:
private:
    treenode<Key,Data> * left;
    treenode<Key,Data> * right;
    treenode<Key,Data> * parent;
    avl_tree_skew skew;
    Key _key;
    Data _data;
    treenode(const Key & __key, const Data & __data):left(NULL),right(NULL),parent(NULL),skew(noone),_key(__key),_data(__data){}
    //operator=

friend ostream& operator<< <Key,Data> ( ostream & o, const treenode<Key,Data> & v);
friend bool operator <  <Key,Data> ( const treenode<Key,Data> & a, const treenode<Key,Data> & b);
friend bool operator >  <Key,Data> ( const treenode<Key,Data> & a, const treenode<Key,Data> & b);
};/*}}}*/ 
template <class Key,class Data>
inline bool operator >  ( const treenode<Key,Data> & a, const treenode<Key,Data> & b)/*{{{*/
{
    return ( a._key > b._key );
}
/*}}}*/
template <class Key,class Data>
inline bool operator <  ( const treenode<Key,Data> & a, const treenode<Key,Data> & b)/*{{{*/
{
    return ( a._key < b._key );
}/*}}}*/
template <class Key,class Data>
inline ostream& operator<< ( ostream & o, const treenode<Key,Data> & v)/*{{{*/
{
    o << "[" << v._key << "|" << v._data << "]";
    return o;
}/*}}}*/
template <class Key,class Data>
class bintreeiterator/*{{{*/
{
private:
    bintree< Key, Data> * tree;
    treenode< Key, Data> * _actual;
    treenode< Key, Data> *const* root;
    
    
    bintreeiterator( ) { }
protected:
public:
    bintreeiterator( const bintree<Key,Data>& bt ) : _actual(bt.Root),root(&bt.Root) { }

    virtual ~bintreeiterator( ) { }
    
    inline bool first( void );

    inline bool valid( void ) const { return (_actual); }

    inline bool next( void );

//    inline bool prev( void );

    inline bool last( void );

    inline Key& key ( void ) const;

    inline Data& data ( void ) const;

};/*}}}*/
template <class Key,class Data>
bool bintreeiterator<Key,Data>::first( void )/*{{{*/
{
    treenode<Key,Data> * ptr = *root;
   
    if (!ptr) 
    {
	return false; 
    }

    while (ptr->left)
    {
	ptr = ptr->left;
    }
    _actual = ptr;
    return true;
}/*}}}*/
template <class Key,class Data>
bool bintreeiterator<Key,Data>::next( void )/*{{{*/
{
    register treenode<Key,Data> * ptr;
    
    if ( _actual->right )
    {
	ptr = _actual->right;
	while ( ptr->left )
	{
	    ptr = ptr->left;
	}
	_actual = ptr;
	return true;
    }
    else if ( _actual->parent )
    {
	while (_actual->parent != NULL)
	{
	    ptr = _actual->parent;
	    if ( ptr->left == _actual )
	    {
		_actual = ptr;
		return true;
	    }
	    _actual = ptr;
	}
	_actual = NULL;
	return false;
    }
    else
    {
	_actual = NULL;
	return false;
    }
    

}/*}}}*/
template <class Key,class Data>
bool bintreeiterator<Key,Data>::last( void )/*{{{*/
{
    treenode<Key,Data> * ptr = *root;

    if ( !ptr ) 
    {
	return false;
    }
    
    while (ptr->right)
    {
	ptr = ptr->right;
    }
    _actual = ptr;
    return true;
}/*}}}*/
template <class Key,class Data>
Key& bintreeiterator<Key,Data>::key( void ) const/*{{{*/
{
    if (_actual)
    {
	return _actual->_key;
    }
    else
    {
	throw x_item("bintreeiterator::key : Actual Item is not valid !!",__FLF__);
    }
}/*}}}*/
template <class Key,class Data>
Data& bintreeiterator<Key,Data>::data( void ) const/*{{{*/
{
    if (_actual)
    {
	return _actual->_data;
    }
    else
    {
	throw x_item("bintreeiterator::key : Actual Item is not valid !!",__FLF__);
    }
}/*}}}*/

/**
 *$Log: bintree.h,v $
 *Revision 1.8  2003/02/25 16:37:28  plasmahh
 *mstring bugfix
 *mstring enhancement
 *
 *Revision 1.7  2003/01/07 13:35:42  plasmahh
 *new tree stuff
 *fixed somethings in mstring
 *begin making better socket classes
 *
 *Revision 1.6  2002/11/25 23:38:24  plasmahh
 *new function for dllist
 *new exception stuff
 *
 *Revision 1.5  2002/11/03 22:21:56  plasmahh
 *fixed some little stuff, and implemented in bintree
 *
 *Revision 1.4  2002/10/31 22:34:46  plasmahh
 *mstring fix
 *
 *Revision 1.3  2002/10/24 08:09:50  plasmahh
 *bintree implementation changes
 *
 *Revision 1.2  2002/10/22 14:01:08  plasmahh
 *implementing binary tree
 *
 *Revision 1.1  2002/10/21 23:48:24  plasmahh
 *New binary tree class
 *
 *Revision 1.1  2002/10/01 21:07:26  plasmahh
 *new math stuff
 *
 *Revision 1.1  2002/06/10 10:36:07  plasmahh
 *Added some math classes in design state, not to be compiled
 *
 */

